//
// Copyright © 2017 Arm Ltd. All rights reserved.
// SPDX-License-Identifier: MIT
//
#pragma once

namespace armnn
{

constexpr const char * NeonBackendId() { return "CpuAcc"; }

} // namespace armnn
