//
// Copyright © 2020 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//
#pragma once

namespace arm
{

namespace pipe
{
    static const unsigned int PIPE_MAGIC = 0x45495434;
} // namespace pipe
} // namespace arm