//
// Copyright © 2022 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//
#include <cstdio>

int main()
{
    printf("The bare metal deserialized graph was successfully executed\n");
    return 0;
}
