//
// Copyright © 2017 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//
#pragma once

#include <Runtime.hpp>

namespace armnn
{

void RuntimeLoadedNetworksReserve(armnn::RuntimeImpl* runtime);

} // namespace armnn
