//
// Copyright © 2017, 2023 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once
#include "NeonAbsWorkload.hpp"
#include "NeonActivationWorkload.hpp"
#include "NeonAdditionWorkload.hpp"
#include "NeonArgMinMaxWorkload.hpp"
#include "NeonBatchMatMulWorkload.hpp"
#include "NeonBatchNormalizationWorkload.hpp"
#include "NeonBatchToSpaceNdWorkload.hpp"
#include "NeonCastWorkload.hpp"
#include "NeonChannelShuffleWorkload.hpp"
#include "NeonComparisonWorkload.hpp"
#include "NeonConcatWorkload.hpp"
#include "NeonConstantWorkload.hpp"
#include "NeonConvertFp16ToFp32Workload.hpp"
#include "NeonConvertFp32ToFp16Workload.hpp"
#include "NeonConvolution2dWorkload.hpp"
#include "NeonConvolution3dWorkload.hpp"
#include "NeonDepthToSpaceWorkload.hpp"
#include "NeonDepthwiseConvolutionWorkload.hpp"
#include "NeonDequantizeWorkload.hpp"
#include "NeonDetectionPostProcessWorkload.hpp"
#include "NeonDivisionWorkload.hpp"
#include "NeonElementwiseBinaryWorkload.hpp"
#include "NeonExpWorkload.hpp"
#include "NeonFillWorkload.hpp"
#include "NeonFloorFloatWorkload.hpp"
#include "NeonFullyConnectedWorkload.hpp"
#include "NeonGatherWorkload.hpp"
#include "NeonGatherNdWorkload.hpp"
#include "NeonInstanceNormalizationWorkload.hpp"
#include "NeonL2NormalizationFloatWorkload.hpp"
#include "NeonLogWorkload.hpp"
#include "NeonLogicalAndWorkload.hpp"
#include "NeonLogicalNotWorkload.hpp"
#include "NeonLogicalOrWorkload.hpp"
#include "NeonLogSoftmaxWorkload.hpp"
#include "NeonLstmFloatWorkload.hpp"
#include "NeonMaximumWorkload.hpp"
#include "NeonMeanWorkload.hpp"
#include "NeonMinimumWorkload.hpp"
#include "NeonMultiplicationWorkload.hpp"
#include "NeonNegWorkload.hpp"
#include "NeonNormalizationFloatWorkload.hpp"
#include "NeonPadWorkload.hpp"
#include "NeonPermuteWorkload.hpp"
#include "NeonPooling2dWorkload.hpp"
#include "NeonPooling3dWorkload.hpp"
#include "NeonPreluWorkload.hpp"
#include "NeonQLstmWorkload.hpp"
#include "NeonQuantizedLstmWorkload.hpp"
#include "NeonQuantizeWorkload.hpp"
#include "NeonRankWorkload.hpp"
#include "NeonReduceWorkload.hpp"
#include "NeonReshapeWorkload.hpp"
#include "NeonResizeWorkload.hpp"
#include "NeonRsqrtWorkload.hpp"
#include "NeonSinWorkload.hpp"
#include "NeonSliceWorkload.hpp"
#include "NeonSoftmaxWorkload.hpp"
#include "NeonSpaceToBatchNdWorkload.hpp"
#include "NeonSpaceToDepthWorkload.hpp"
#include "NeonSplitterWorkload.hpp"
#include "NeonSqrtWorkload.hpp"
#include "NeonStackWorkload.hpp"
#include "NeonStridedSliceWorkload.hpp"
#include "NeonSubtractionWorkload.hpp"
#include "NeonTileWorkload.hpp"
#include "NeonTransposeConvolution2dWorkload.hpp"
#include "NeonTransposeWorkload.hpp"
#include "NeonUnidirectionalSequenceLstmFloatWorkload.hpp"
#include "NeonUnidirectionalSequenceLstmWorkload.hpp"