//
// Copyright © 2021 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once

#include <armnn/Types.hpp>

namespace armnnUtils
{

template<typename T>
bool CompatibleTypes(armnn::DataType);

} //namespace armnnUtils
