//
// Copyright © 2020 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once

namespace arm
{
namespace pipe
{

int GetCurrentProcessId();

} // namespace arm
} // namespace pipe
