//
// Copyright © 2021 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once

namespace armnn
{

class ConstTensorHandle;
class TensorHandle;
class ScopedTensorHandle;
class PassthroughTensorHandle;
class ConstPassthroughTensorHandle;

} // namespace armnn
