//
// Copyright © 2022 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once

namespace armnn
{

namespace experimental
{

struct ExecutionData
{
    void* m_Data;
};

} // end experimental namespace

} // end armnn namespace
