//
// Copyright © 2022 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//
#pragma once

namespace armnn
{

constexpr const char * TosaRefBackendId() { return "TosaRef"; }

} // namespace armnn
