//
// Copyright © 2021-2023 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#include <armnn/delegate/classic/include/Version.hpp>
#pragma message("armnn/delegate/include/Version.hpp has been deprecated, it is due for removal in"\
                " 24.02 release. Please include armnn/delegate/classic/include/Version.hpp instead.")
