//
// Copyright © 2017,2023 Arm Ltd and Contributors. All rights reserved.
// SPDX-License-Identifier: MIT
//

#pragma once
#include "ClAbsWorkload.hpp"
#include "ClActivationWorkload.hpp"
#include "ClAdditionWorkload.hpp"
#include "ClArgMinMaxWorkload.hpp"
#include "ClComparisonWorkload.hpp"
#include "ClConstantWorkload.hpp"
#include "ClBatchMatMulWorkload.hpp"
#include "ClBatchNormalizationFloatWorkload.hpp"
#include "ClBatchToSpaceNdWorkload.hpp"
#include "ClCastWorkload.hpp"
#include "ClChannelShuffleWorkload.hpp"
#include "ClConvolution2dWorkload.hpp"
#include "ClConvolution3dWorkload.hpp"
#include "ClDepthToSpaceWorkload.hpp"
#include "ClDepthwiseConvolutionWorkload.hpp"
#include "ClDequantizeWorkload.hpp"
#include "ClDivisionWorkload.hpp"
#include "ClElementwiseBinaryWorkload.hpp"
#include "ClExpWorkload.hpp"
#include "ClFillWorkload.hpp"
#include "ClFloorFloatWorkload.hpp"
#include "ClFullyConnectedWorkload.hpp"
#include "ClGatherWorkload.hpp"
#include "ClGatherNdWorkload.hpp"
#include "ClInstanceNormalizationWorkload.hpp"
#include "ClL2NormalizationFloatWorkload.hpp"
#include "ClLogWorkload.hpp"
#include "ClLogicalAndWorkload.hpp"
#include "ClLogicalNotWorkload.hpp"
#include "ClLogicalOrWorkload.hpp"
#include "ClLogSoftmaxWorkload.hpp"
#include "ClLstmFloatWorkload.hpp"
#include "ClConcatWorkload.hpp"
#include "ClMaximumWorkload.hpp"
#include "ClMeanWorkload.hpp"
#include "ClMinimumWorkload.hpp"
#include "ClMultiplicationWorkload.hpp"
#include "ClNegWorkload.hpp"
#include "ClNormalizationFloatWorkload.hpp"
#include "ClPermuteWorkload.hpp"
#include "ClPadWorkload.hpp"
#include "ClPooling2dWorkload.hpp"
#include "ClPooling3dWorkload.hpp"
#include "ClPreluWorkload.hpp"
#include "ClQLstmWorkload.hpp"
#include "ClQuantizeWorkload.hpp"
#include "ClQuantizedLstmWorkload.hpp"
#include "ClRankWorkload.hpp"
#include "ClReduceWorkload.hpp"
#include "ClReshapeWorkload.hpp"
#include "ClResizeWorkload.hpp"
#include "ClRsqrtWorkload.hpp"
#include "ClSinWorkload.hpp"
#include "ClSliceWorkload.hpp"
#include "ClSoftmaxWorkload.hpp"
#include "ClSpaceToBatchNdWorkload.hpp"
#include "ClSpaceToDepthWorkload.hpp"
#include "ClSplitterWorkload.hpp"
#include "ClSqrtWorkload.hpp"
#include "ClStackWorkload.hpp"
#include "ClStridedSliceWorkload.hpp"
#include "ClSubtractionWorkload.hpp"
#include "ClConvertFp16ToFp32Workload.hpp"
#include "ClConvertFp32ToFp16Workload.hpp"
#include "ClTileWorkload.hpp"
#include "ClTransposeConvolution2dWorkload.hpp"
#include "ClTransposeWorkload.hpp"
#include "ClUnidirectionalSequenceLstmFloatWorkload.hpp"
